#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include "ieee754.h"
#include "coeff.h"
#include "common.h"

extern const struct_coeff	coeff_burrbrown;
extern const struct_coeff	coeff_DAC7_1st;
extern const struct_coeff	coeff_DAC7_2nd;
extern const struct_coeff	coeff_DAC7_3rd;
extern const struct_coeff	coeff_FDA_44k;
extern const struct_coeff	coeff_HUGO;

#define	TRUE		(-1)			// boolean like value
#define	FALSE		0			// boolean like value
#define	MSB_32BIT	0x80000000		// only bit 31 is on
#define	max(a,b)	(((a) > (b)) ? (a) : (b))

enum e_filter_kind {
	E_burrbrown		= 0,
	E_burrbrown_sparse	= 1,
	E_DAC7_1st		= 2,
	E_DAC7_2nd		= 3,
	E_DAC7_3rd		= 4,
	E_FDA_44k		= 5,
	E_HUGO			= 6,
	E_FILTER_COUNT
};
static const struct_coeff	*struct_coeffs[] = {
	&coeff_burrbrown,
	&coeff_burrbrown,
	&coeff_DAC7_1st,
	&coeff_DAC7_2nd,
	&coeff_DAC7_3rd,
	&coeff_FDA_44k,
	&coeff_HUGO
};

struct s_fcoeff {
	int	length;
	int	bit_resolution;
	int	data_resolution;
	float	fscale;
	float	*coeff;
	int	data_index;
	float	*data;
}	s_fcoeffs[E_FILTER_COUNT];

static int	copy_filter(const struct_coeff *p_sc,
				 struct s_fcoeff *ps_fcoeffp)
{
	const int	length = p_sc->length;
	int		i;

	ps_fcoeffp->length = length;
	ps_fcoeffp->coeff = malloc(length * sizeof(float));
	ps_fcoeffp->data = malloc(length * sizeof(float));
	if (ps_fcoeffp->coeff == NULL || ps_fcoeffp->data == NULL) {
		fprintf(stderr, "malloc failed\n");
		return FALSE;
	}

	memcpy(ps_fcoeffp->coeff, p_sc->filter, length * sizeof(float));
	// get max value of filter
	ps_fcoeffp->fscale = 0.0f;
	for (i = 0; i < length; i++) {
		const float	current = fabs(ps_fcoeffp->coeff[i]);
		// set float coeff filter
		if (ps_fcoeffp->fscale < current) {
			ps_fcoeffp->fscale = current;
		}
	}
	ps_fcoeffp->fscale *= 1.10f;

	// normalize filter value (set ps_fcoeffp->coeff value to 1.0f or -1.0f)
	for (i = 0; i < length; i++) {
		ps_fcoeffp->coeff[i] /= ps_fcoeffp->fscale;
		strong_limitbits(ps_fcoeffp->coeff + i,
				 ps_fcoeffp->bit_resolution);
		// clear data buffer
		ps_fcoeffp->data[i] = 0.0f;
	}
	ps_fcoeffp->data_index = 0;
	return TRUE;
}

static inline void	limit_resolution(enum e_filter_kind e_filter,
					int limit_filter, int limit_data)
{
	if (s_fcoeffs[e_filter].data_resolution > limit_data) {
		s_fcoeffs[e_filter].data_resolution = limit_data;
	}
	if (s_fcoeffs[e_filter].bit_resolution > limit_filter) {
		s_fcoeffs[e_filter].bit_resolution = limit_filter;
	}
}

int	init_all_filters(int data_resolution)
{
	int	i;

	for (i = 0; i < E_FILTER_COUNT; i++) {
		s_fcoeffs[i].data_resolution = 32;
		s_fcoeffs[i].bit_resolution = 32;
	}

	limit_resolution(E_burrbrown, 16, 16);	// I assumed 16x16 bit calc.
	limit_resolution(E_burrbrown_sparse, 16, 16);
	limit_resolution(E_DAC7_1st, 22, 20);
	limit_resolution(E_DAC7_2nd, 22, 20);
	limit_resolution(E_DAC7_3rd, 22, 20);
	limit_resolution(E_HUGO, 18, 18);	// spartan6 FPGA module spec.

	for (i = 0; i < E_FILTER_COUNT; i++) {
		if ( !copy_filter(struct_coeffs[i], s_fcoeffs + i) ) {
			goto err_exit;
		}
	}

	return TRUE;

err_exit:
	do {
		if (s_fcoeffs[i].coeff) {
			free(s_fcoeffs[i].coeff);
			s_fcoeffs[i].coeff = NULL;
		}
		if (s_fcoeffs[i].data) {
			free(s_fcoeffs[i].data);
			s_fcoeffs[i].data = NULL;
		}
	} while (--i >= 0);
	return FALSE;
}

static float	do_coeff(struct s_fcoeff *p_coeff, float data,
			 int input_resolution)
{
	int	i;
	int	j;
	int	max_resolution;
	float	sum = 0.0f;

	if (p_coeff->data_resolution < input_resolution) {
		input_resolution = p_coeff->data_resolution;
	}
	max_resolution = max(p_coeff->bit_resolution, input_resolution);
	p_coeff->data[p_coeff->data_index] = data;

	if (input_resolution < 32) {
		strong_limitbits(p_coeff->data + p_coeff->data_index,
				  input_resolution);
	}

	p_coeff->data_index++;
	p_coeff->data_index %= p_coeff->length;
	for (i = 0, j = p_coeff->data_index; i < p_coeff->length; ) {
		sum += (p_coeff->coeff[i++] * p_coeff->data[j++]);
		if (j >= p_coeff->length) {
			j = 0;
		}
	}
	sum *= p_coeff->fscale;
	strong_limitbits(&sum, max_resolution);
	if (sum > 1.0f) {
		printf("over flow %f\n", sum);
		sum = 1.0f;
	}
	if (sum < -1.0f) {
		printf("under flow %f\n", sum);
		sum = -1.0f;
	}
	return sum;
}

float	*do_burrbrown_coeff(float data, int resolution)
{
	int		i;
	static float	result[OVERSAMPLING_RATE * 2];

	for (i = 0; i < OVERSAMPLING_RATE; i++) {
		result[i] = do_coeff(s_fcoeffs + E_burrbrown, data, resolution);
	}

	result[OVERSAMPLING_RATE] =
		 do_coeff(s_fcoeffs + E_burrbrown_sparse, data, resolution);
	for (i = 1; i < OVERSAMPLING_RATE; i++) {
		result[i + OVERSAMPLING_RATE] =
		 do_coeff(s_fcoeffs + E_burrbrown_sparse, 0.0f,
					 resolution);
	}
	return result;
}

float	*do_FDA_44k_coeff(float data, int resolution)
{
	int		i;
	static float	result[OVERSAMPLING_RATE];

	for (i = 0; i < OVERSAMPLING_RATE; i++) {
		result[i] = do_coeff(s_fcoeffs + E_FDA_44k, data, resolution);
	}
	return result;
}

float	*do_HUGO_coeff(float data, int resolution)
{
	int		i;
	static float	result[OVERSAMPLING_RATE];

	for (i = 0; i < OVERSAMPLING_RATE; i++) {
		result[i] = do_coeff(s_fcoeffs + E_HUGO, data, resolution);
	}
	return result;
}

float	*do_DAC7_coeff(float data, int resolution)
{
	int		i;
	static float	result[OVERSAMPLING_RATE];
	float		result2 = 0.0f;
	float		result4 = 0.0f;

	for (i = 0; i < OVERSAMPLING_RATE; i++) {
		if ( (i & 3) == 0) {
			result2 = do_coeff(s_fcoeffs + E_DAC7_1st,
						 data, resolution);
		}
		if ( (i & 1) == 0) {
			result4 = do_coeff(s_fcoeffs + E_DAC7_2nd,
						 result2, resolution);
		}
		result[i] = do_coeff(s_fcoeffs + E_DAC7_3rd,
						 result4, resolution);
	}
	return result;
}

static float	search_max(float *pf, int len)
{
	float max = 0.0f;

	while (len-- > 0) {
		const float	f = fabs(*pf++);
		if (f > max) {
			max = f;
		}
	}
	return max;
}

void	HUGO_estimate(void)
{
	float	outer;
	float	inner;

	outer = search_max(s_fcoeffs[E_HUGO].coeff,
		 (s_fcoeffs[E_HUGO].length - s_fcoeffs[E_FDA_44k].length) / 2);
	inner = search_max(s_fcoeffs[E_HUGO].coeff +
		 (s_fcoeffs[E_HUGO].length - s_fcoeffs[E_FDA_44k].length) / 2,
		 s_fcoeffs[E_FDA_44k].length / 2);

	printf("%f\n", outer / inner);
	printf("%f bits smaller\n", logf(outer / inner) / logf(2.0f));
}
