#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "common.h"
#include "ieee754.h"

static const char *file_names[] = {
	"burrbrown.dat",
	"burrbrown_sparse.dat",
	"dac7.dat",
	"fda.dat",
	"fda_pwm.dat",
	"fda_pwm_nons.dat",
	"fda_pwm_3level.dat",
	"fda_nos.dat",
	"hugo.dat" };

#define	CHECKPOINT		printf("%s() %d\n", __func__, __LINE__)
#define	OUT_SAMPLING_HZ		352800
#define	SAMPLING_HZ		44100
#define	SIGNAL_HZ		((double)SAMPLING_HZ / 256.0)
#define	DURATION		2
#define	OUT_SCALE		0.9
#define	CYCLE			32

#define	PWM_RESOLUTION			32
#define	PWM_RESOLUTION_FLOAT		32.0f
#define	PWM_3LEVEL_RESOLUTION		64
#define	PWM_3LEVEL_RESOLUTION_FLOAT	64.0f
#define	NOS_RESOLUTION			2048
#define	NOS_RESOLUTION_FLOAT		2048.0f

extern int	init_all_filters(int);
extern float	*do_burrbrown_coeff(float data, int resolution);
extern float	*do_DAC7_coeff(float data, int resolution);
extern float	*do_FDA_44k_coeff(float data, int resolution);
extern float	*do_HUGO_coeff(float data, int resolution);

int	main(int argc, char *argv[])
{
	FILE	*fps[E_METHOD_COUNT];
	int	i;
	int	j;
	int	data_resolution = 32;	// user full bits
	float	signals[CYCLE];
	float	sum_pwm = 0.0f;
	float	sum_pwm_3level = 0.0f;
	float	sum_nos = 0.0f;
	float	pwm_buff[OVERSAMPLING_RATE];
	float	pwm_3level_buff[OVERSAMPLING_RATE];
	float	pwm_nons_buff[OVERSAMPLING_RATE];

	for (i = 1; i < argc; i++) {
		if (memcmp(argv[i], "-bit=", 5) == 0) {
			sscanf(argv[i], "-bit=%d", &data_resolution);
		} else {
			fprintf(stderr, "unknown option %s\n", argv[i]);
			return 1;
		}
	}

	if ( !init_all_filters(data_resolution) ) {
		fprintf(stderr, "filter initialize failed.\n");
		exit(1);
	}
	for (i = 0; i < E_METHOD_COUNT; i++) {
		fps[i] = fopen(file_names[i], "w");
		if (fps[i] == NULL) {
			fprintf(stderr,
				"file %s open failed.\n", file_names[i]);
			exit(1);
		}
	}

	// calculate sin table
	for (i = 0; i < CYCLE; i++) {
		const double	theta = 2 * M_PI * (i % CYCLE) / (float)CYCLE;
		signals[i] = (float)(OUT_SCALE * sin(theta));
		strong_limitbits(signals + i, data_resolution);
	}
	// calculate Low-Pass FIR filters' output
	for (i = 0; i < SAMPLING_HZ * DURATION; i++) {
		const float	signal = signals[i % CYCLE];
		float		*p_result;
		int		pwm;

		printf("\r  calcurating %d/%d", i, SAMPLING_HZ * DURATION);
		fflush(stdout);
		p_result = do_burrbrown_coeff(signal, data_resolution);
		fwrite(p_result, sizeof(float), OVERSAMPLING_RATE, 
					fps[E_METHOD_BURRBROWN]);
		fwrite(p_result + OVERSAMPLING_RATE, sizeof(float),
			 OVERSAMPLING_RATE, fps[E_METHOD_BURRBROWN_SPARSE]);
		p_result = do_DAC7_coeff(signal, data_resolution);
		fwrite(p_result, sizeof(float), OVERSAMPLING_RATE, 
					fps[E_METHOD_DAC7]);
		p_result = do_FDA_44k_coeff(signal, data_resolution);
		fwrite(p_result, sizeof(float), OVERSAMPLING_RATE, 
					fps[E_METHOD_FDA]);
		for (j = 0; j < OVERSAMPLING_RATE; j++) {
			sum_pwm += p_result[j];
			if (sum_pwm >= 1.0f) {
				pwm = PWM_RESOLUTION;
			} else if (sum_pwm <= -1.0) {
				pwm = -PWM_RESOLUTION;
			} else if (sum_pwm >= 0.0f) {
				pwm = roundf(sum_pwm * PWM_RESOLUTION_FLOAT);
			} else {
				pwm = -roundf(-sum_pwm * PWM_RESOLUTION_FLOAT);
			}
			sum_pwm -= pwm / PWM_RESOLUTION_FLOAT;
			pwm_buff[j] = pwm / PWM_RESOLUTION_FLOAT;

			sum_pwm_3level += p_result[j];
			if (sum_pwm_3level >= 1.0f) {
				pwm = PWM_3LEVEL_RESOLUTION;
			} else if (sum_pwm_3level <= -1.0) {
				pwm = -PWM_3LEVEL_RESOLUTION;
			} else if (sum_pwm_3level >= 0.0f) {
				pwm = roundf(sum_pwm_3level * 
					PWM_3LEVEL_RESOLUTION_FLOAT);
			} else {
				pwm = -roundf(-sum_pwm_3level * 
					PWM_3LEVEL_RESOLUTION_FLOAT);
			}
			sum_pwm_3level -= pwm / PWM_3LEVEL_RESOLUTION_FLOAT;
			pwm_3level_buff[j] = pwm / PWM_3LEVEL_RESOLUTION_FLOAT;

			pwm_nons_buff[j] = p_result[j];
			if (pwm_nons_buff[j] >= 1.0f) {
				pwm_nons_buff[j] = 1.0f;
			} else if (pwm_nons_buff[j] <= -1.0) {
				pwm_nons_buff[j] = -1.0f;
			} else if (pwm_nons_buff[j] >= 0.0f) {
				pwm_nons_buff[j] = roundf(pwm_nons_buff[j] * 
					PWM_RESOLUTION_FLOAT) / 
					PWM_RESOLUTION_FLOAT;
			} else {
				pwm_nons_buff[j] = -roundf(-pwm_nons_buff[j] * 
					PWM_RESOLUTION_FLOAT) / 
					PWM_RESOLUTION_FLOAT;
			}
		}
		if (fwrite(pwm_buff, sizeof(float), OVERSAMPLING_RATE, 
				fps[E_METHOD_FDA_PWM]) != OVERSAMPLING_RATE) {
			fprintf(stderr, "fwrite(%s) failed\n",
				 file_names[E_METHOD_FDA_PWM]);
		}
		if (fwrite(pwm_3level_buff, sizeof(float), OVERSAMPLING_RATE, 
			fps[E_METHOD_FDA_PWM_3LEVEL]) != OVERSAMPLING_RATE) {
			fprintf(stderr, "fwrite(%s) failed\n",
				 file_names[E_METHOD_FDA_PWM_3LEVEL]);
		}
		if (fwrite(pwm_nons_buff, sizeof(float), OVERSAMPLING_RATE, 
			   fps[E_METHOD_FDA_PWM_NONS]) != OVERSAMPLING_RATE) {
			fprintf(stderr, "fwrite(%s) failed\n",
				 file_names[E_METHOD_FDA_PWM_NONS]);
		}
		for (j = 0; j < OVERSAMPLING_RATE; j++) {
			sum_nos += p_result[j];
			if (sum_nos >= 0.0f) {
				pwm = roundf(sum_nos * NOS_RESOLUTION_FLOAT);
				if (pwm > (NOS_RESOLUTION - 1)) {
					pwm = (NOS_RESOLUTION - 1);
				}
			} else {
				pwm = -roundf(-sum_nos * NOS_RESOLUTION_FLOAT);
				if (pwm < -NOS_RESOLUTION) {
					pwm = -NOS_RESOLUTION;
				}
			}
			sum_nos -= pwm / NOS_RESOLUTION_FLOAT;
			pwm_buff[j] = pwm / NOS_RESOLUTION_FLOAT;
		}
		if (fwrite(pwm_buff, sizeof(float), OVERSAMPLING_RATE, 
				fps[E_METHOD_FDA_NOS]) != OVERSAMPLING_RATE) {
			fprintf(stderr, "fwrite(%s) failed\n",
				 file_names[E_METHOD_FDA_NOS]);
		}
		p_result = do_HUGO_coeff(signal, data_resolution);
		fwrite(p_result, sizeof(float), OVERSAMPLING_RATE, 
					fps[E_METHOD_HUGO]);
	}
	printf("\r  finished                              \n");
	for (i = 0; i < E_METHOD_COUNT; i++) {
		fclose(fps[i]);
	}
	return 0;
}
